from __future__ import annotations
from typing import IO, Any, Callable, Iterator, Sequence
from typing_extensions import Iterable, Type, TypeVar


class Renderable:
    priority: int = 0
    associativity = "both"

    def stream(self, context: Context) -> TokenStream:
        return NotImplemented

    def wrap(self) -> Renderable:
        """
        This method should return a version of Renderable that retains the same meaning but has higher priority.
        For example, if your renderable is a Python expression `a + b`, this method should return `(a + b)`

        By default, it wraps in parens `x` -> `(x)`
        """
        return Parens(self)

    def priority_comparator(
        self,
        operation: Renderable,
        side: str = "none",
    ) -> bool:
        """Used to decide if the expression should be wrapped to preserve priority and associativity rules."""
        if self.priority < operation.priority:
            return True

        if self.priority == operation.priority:
            if side == "none" or self.associativity == "both":
                return False
            if self.associativity != side:  # e.g "left" != "right" and vice versa
                return True

        return False

    def respect_priority(
        self,
        operation: Renderable,
        comparator: Comparator | None = None,
        side: str = "none",
    ) -> Renderable:
        """
        This method should be called to avoid breaking priority in complex expressions.

        It should be called from parent expression __init__

        `operation` is the object of parent expression (operation)
        `comparator` is a function that should return True if the expression should be wrapped.
        If no comparator is passed, `self.priority_comparator` is used as comparator.

        `side` is a parameter denoting the side of `self` in the expression to support associativity (or any other side-specific behaviour).
        This is mostly arbitrary, but the basic convention is to use "left" and "right" for binary expression, and "none" as default.
        """

        if not comparator:

            def default_comparator(self: Renderable, operation: Renderable, side: str):
                return self.priority_comparator(operation, side)

            comparator = default_comparator

        if comparator(self, operation, side):
            return self.wrap()

        return self

    def render_string(self, context: Payload | None = None):
        """Render component with a provided context. Check `Renderer` class for advanced rendering"""
        return Renderer(context).render_string(self)


class Marker:
    def __init__(self, key: str, data: Payload):
        self.key = key
        self.data = {k: data[k] for k in sorted(data)}

    def __hash__(self) -> int:
        return hash((self.key, tuple(self.data.items())))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Marker):
            return False

        return self.key == other.key and self.data == other.data


_T = TypeVar("_T", str, bytes)


class StreamFile(IO[_T]):
    def __init__(self, stream: Iterable[_T], cls: Type[_T]):
        self.stream: Iterator[_T] = iter(stream)
        self.empty: _T = cls()
        self.buffer: _T = self.empty

    def get_next(self) -> _T | None:
        if self.buffer:
            return self.store()
        return next(self.stream, None)

    def store(self, item: _T | None = None) -> _T:
        prev = self.buffer

        if item is None:
            item = self.empty

        self.buffer = item
        return prev

    def join(self, buf: Iterable[_T]) -> _T:
        return self.empty.join(buf)

    def read(self, __n: int = -1) -> _T:
        local_buffer: list[_T] = []
        remaining = __n

        if __n < 0:
            return self.join(self.stream)

        if not __n:
            return self.empty

        while remaining >= 0:
            elem = self.get_next()

            if elem is None:
                break

            if not elem:
                continue  # no need to join empty strings

            elem_length = len(elem)

            if elem_length < remaining:
                local_buffer.append(elem)
            else:
                local_buffer.append(elem[:remaining])
                self.store(elem[remaining:])
                break

            remaining -= elem_length

        return self.join(local_buffer)

    def readlines(self, __hint: int = -1) -> list[_T]:
        buf: list[_T] = []

        for _ in range(__hint):
            line = self.readline()
            if not line:
                break

        return buf

    def readline(self, __limit: int = -1) -> _T:
        return super().readline(__limit)


class Renderer:
    def __init__(self, context: Payload | None = None):
        self.context = context or {}

    def get_subrenderer(self, local_context: Payload) -> Renderer:
        return Renderer(
            {
                **self.context,
                **local_context,
            }
        )

    def stream(
        self, renderable: Renderable, local_context: Payload | None = None
    ) -> TokenStream:
        if local_context:
            sub_renderer = self.get_subrenderer(local_context)
            return sub_renderer.stream(renderable)

        return renderable.stream(Context(self))

    def byte_stream(
        self, renderable: Renderable, local_context: Payload | None = None
    ) -> ByteStream:
        for token in self.str_stream(renderable, local_context):
            yield token.encode("utf-8")

    def str_stream(
        self, renderable: Renderable, local_context: Payload | None = None
    ) -> StrStream:
        for token in self.stream(renderable, local_context):
            if isinstance(token, str):
                yield token

    def render_bytes(
        self, renderable: Renderable, local_context: Payload | None = None
    ):
        return b"".join(
            self.byte_stream(
                renderable,
                local_context,
            )
        )

    def render_string(
        self, renderable: Renderable, local_context: Payload | None = None
    ) -> str:
        return "".join(
            self.str_stream(
                renderable,
                local_context,
            )
        )

    def as_file(
        self, renderable: Renderable, local_context: Payload | None = None
    ) -> IO[str]:
        return StreamFile(
            self.str_stream(
                renderable,
                local_context,
            ),
            str,
        )

    def as_binary_file(
        self, renderable: Renderable, local_context: Payload | None = None
    ) -> IO[bytes]:
        return StreamFile(
            self.byte_stream(
                renderable,
                local_context,
            ),
            bytes,
        )


class Context:
    def __init__(self, renderer: Renderer):
        self._renderer = renderer

    def __getattr__(self, __name: str) -> Any:
        return self._renderer.context.get(__name)

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name == "_renderer":
            return super().__setattr__(__name, __value)
        self._renderer.context[__name] = __value

    def derive(self, **kwargs: Any):
        return Context(
            self._renderer.get_subrenderer(kwargs),
        )


def get_render(context: Context) -> Callable[[Renderable], TokenStream]:
    return lambda r: r.stream(context)


from .internal_types import (
    ByteStream,
    Comparator,
    Payload,
    StrStream,
    TokenStream,
)

from .startkit import Parens

_RT = TypeVar("_RT", bound=Renderable)



from __future__ import annotations

from .stream_utils import add_tab, separated
from .core import Context, Renderable, get_render
from .internal_types import StreamTransformer, Token, TokenStream


class Parens(Renderable):
    """Generic wrapper"""

    def __init__(
        self, contents: Renderable, open_paren: Token = "(", close_paren: Token = ")"
    ):
        self.contents = contents
        self.parens = (open_paren, close_paren)

    def wrap(self):
        return self

    def render_parts(self, context: Context) -> TokenStream:
        yield from self.contents.stream(context)

    def stream(self, context: Context) -> TokenStream:
        yield self.parens[0]
        yield from self.render_parts(context)
        yield self.parens[1]


class Separated(Renderable):
    """Base block to render a sequence of renderables separated by some separator"""

    def __init__(
        self, *renderables: Renderable, separator: TokenStream, trail: bool = False
    ):
        self.renderables = renderables
        self.separator = separator
        self.trail = trail

    def stream(self, context: Context) -> TokenStream:
        render = get_render(context)
        return separated(
            *map(render, self.renderables),
            separator=self.separator,
            trail=self.trail,
        )


class Block(Renderable):
    def __init__(
        self,
        head: Renderable | None,
        body: Renderable,
        wrapper: StreamTransformer | None = None,
        indenter: StreamTransformer | None = None,
    ) -> None:
        self.head = head
        self.body = body
        self.wrapper: StreamTransformer = wrapper or self.default_wrap
        self.indenter: StreamTransformer = indenter or add_tab

    def default_wrap(self, stream: TokenStream) -> TokenStream:
        yield "{"
        yield " "
        yield "\n"
        yield from stream
        yield "\n"
        yield "}"

    def stream(self, context: Context):
        if self.head:
            yield from self.head.stream(context)

        yield from self.wrapper(self.indenter(self.body.stream(context)))


class Stringify(Renderable):
    def __init__(self, contents: object):
        self.contents = contents

    def stream(self, context: Context) -> TokenStream:
        yield str(self.contents)



from __future__ import annotations
from collections import deque


def add_tab(
    stream: Stream[Token], tab_char: str = "    ", newlines: set[Token] | None = None
) -> Stream[Token]:
    """

    Adds `tab_char` after any token that is equal to any of the newline tokens in `newlines`.

    By default, `tab_char` is "    " (4 ASCII spaces U+0020) and `newlines` is {"\\n"}

    """
    newlines = get_default(newlines, {"\n"})

    yield tab_char

    for token in stream:
        yield token

        if token in newlines:
            yield tab_char


def separated(
    *streams: Stream[Piece], separator: Stream[Piece], trail: bool = False
) -> Stream[Piece]:
    """
    Interjects `separator` tokens between streams, outputting a new chained stream.

    `trail` defines if separator should be added after last token, i.e. if a trailing comma should be inserted
    """

    separator_list = list(
        separator
    )  # separator needs to be stored because it could be a one-time stream

    for i, stream in enumerate(streams):
        yield from stream
        if trail or i + 1 != len(streams):
            yield from separator_list


def prepend(stream: Stream[Piece], *pieces: Piece) -> Stream[Piece]:
    """Injects tokens before stream"""
    return concat(pieces, stream)


def append(stream: Stream[Piece], *pieces: Piece) -> Stream[Piece]:
    """Injects tokens after stream"""
    return concat(stream, pieces)


def omit_start(stream: Stream[Piece], count: int = 1) -> Stream[Piece]:
    """Omits `count` tokens from the start of the stream"""
    for token in stream:
        if count > 0:
            count -= 1
            continue

        yield token


def omit_end(stream: Stream[Piece], count: int = 1) -> Stream[Piece]:
    """Omits `count` tokens from the end of the stream"""
    buf: deque[Piece] = deque()

    for token in stream:
        buf.append(token)

        if len(buf) > count:
            yield buf.popleft()


def concat(*streams: Stream[Piece] | None) -> Stream[Piece]:
    """Concatenates several streams"""
    for stream in streams:
        if not stream:
            continue
        yield from stream


def wrap(
    stream: Stream[Piece],
    prefix: Stream[Piece],
    postfix: Stream[Piece],
) -> Stream[Piece]:
    """Prepends `prefix` before stream and appends `postfix` after stream"""

    return concat(prefix, stream, postfix)


def stream_noop(stream: Stream[Piece]) -> Stream[Piece]:
    return stream


from .internal_types import Piece, Stream, Token
from .utils import get_default



from __future__ import annotations
from typing import Callable, Concatenate, ParamSpec

from .core import Renderable

FuncArgs = ParamSpec("FuncArgs")


class FunctionalRenderable(Renderable):
    def __init__(self, renderer: Callable[[Context], TokenStream]):
        self.renderer = renderer

    def stream(self, context: Context) -> TokenStream:
        yield from self.renderer(context)


def make_renderable(func: Callable[Concatenate[Context, FuncArgs], TokenStream]):
    def call(*args: FuncArgs.args, **kwargs: FuncArgs.kwargs) -> FunctionalRenderable:
        renderable = FunctionalRenderable(
            lambda context: func(context, *args, **kwargs)
        )
        return renderable

    return call


from .core import Context
from .internal_types import TokenStream



from __future__ import annotations
from typing import TypeVar
from typing_extensions import TypeGuard


_T = TypeVar("_T")


def get_default(cond: _T | None, default: _T, value: _T | None = None) -> _T:
    """
    Returns default if cond is None.
    If cond is not None, returns value if it's also not None, else returns cond.

    Two-arg form
    ```python
    get_default(cond, default)
    ```
    is equivalent to

    ```python
    default if cond is None else cond
    ```

    Three-arg form
    ```python
    get_default(cond, default, value)
    ```
    is equivalent to

    ```python
    default if cond is None else value
    ```
    """

    if cond is None:
        return default

    return get_default(value, cond)


def is_marker(token: Token) -> TypeGuard[Marker]:
    return isinstance(token, Marker)


from .core import Marker
from .internal_types import Token



from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Tuple,
    TypeVar,
    Union,
)


Piece = TypeVar("Piece", bound=Union[bytes, str, "Token"])

Token = Union[str, "Marker"]

Stream = Iterable[Piece]
TokenStream = Stream[Token]
StrStream = Stream[str]
ByteStream = Stream[bytes]

StreamTransformer = Callable[[Stream], Stream]

Payload = Dict[str, Any]

Comparator = Callable[["Renderable", "Renderable", str], bool]
ParenSet = Tuple[Token, Token]

from .core import Marker, Renderable
